package processor;

import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.util.Stack;

import javax.swing.JFrame;
import javax.swing.SwingUtilities;

import processor.Action;
import processor.ProcessController;
import graph.LayeredGraphNode;
import processor.Memory.Visibility;
import processor.StackFrame.FrameType;

/**
 * this class executes {@link CodeLine}s and manages the {@link Memory} for an algorithm
 * @author kolja
 *
 */
public class PseudoCodeProcessor extends Thread {

    /**
     * the state of the execution after a step has been done.
     * for example BREAKPOINT indicates that a breakpoint has been reached.
     * @author kolja
     *
     */
    public static enum CodeStatus
    {
        UNFINISHED,
        BREAKPOINT,
        FINISHED
    }

    private LayeredGraphNode graph;
    private String currentDebugOutput;
    private Memory mem;
    protected PseudoCodeNode programPointer;
    protected Stack<ControlFlow> controlStack;
    private boolean skip = false;
    private boolean renderImage = true;
    private JFrame view;
    private ProcessController controller;
    private boolean insideNotSkipLoop = false; // needed to avoid stack overflow errors
    
    /**
     * creates a new {@link PseudoCodeProcessor}
     * @param start the starting line of the algorithm
     * @param graph the graph to work with
     * @param view the view where to display the current state
     */
    public PseudoCodeProcessor( PseudoCodeNode start, LayeredGraphNode graph, JFrame view )
    {
        mem = new Memory();
        mem.addFrame( new StackFrame( FrameType.FUNCTION ) );
        mem.declare( "graph", graph, Visibility.GLOBAL );
        programPointer = start;
        currentDebugOutput = "";
        controlStack = new Stack<>();
        this.graph = graph;
        controller = new ProcessController();
        this.view = view;
    }
    
    public ProcessController getController()
    {
        return controller;
    }
    
    private CodeStatus selectNextNode( PseudoCodeNode next, PseudoCodeNode last )
    {
        programPointer = next;
        last.setSelected( false );
        switch( next.setSelected( true ) ) // check which node should be executed next
        { 
        case CONTINUE:
            // just do the next step
            skip = false;
            return CodeStatus.UNFINISHED;
        case SKIP:
            // node is folded so it is skipped
            skip = true;
            if( insideNotSkipLoop )
                return CodeStatus.UNFINISHED;
            return forwardStepOverUntilNotSkip();
        case STOP:
            // a breakpoint was reached
            skip = false;
            return CodeStatus.BREAKPOINT;
        default:
            return CodeStatus.UNFINISHED;
        }
    }
    
    private CodeStatus selectBeforeNode( PseudoCodeNode next, PseudoCodeNode last )
    {
        programPointer = next;
        last.setSelected( false );
        switch( next.setSelected( true ) ) // check which node should be executed next
        {
        case CONTINUE:
            // just do the previous step
            skip = false;
            return CodeStatus.UNFINISHED;
        case SKIP:
            // node is folded so it is skipped
            skip = true;
            if( insideNotSkipLoop )
                return CodeStatus.UNFINISHED;
            return backwardStepOverUntilNotSkip();
        case STOP:
            // a breakpoint was reached
            skip = false;
            return CodeStatus.BREAKPOINT;
        default:
            return CodeStatus.UNFINISHED;
        }
    }
    
    /**
     * perform one forward step of the algorithm assigned to this processor
     * @return the status of execution after that step
     */
    protected CodeStatus forwardStep()
    {
        // is the program pointer still available?
        if( programPointer == null )
            return CodeStatus.FINISHED;
        
        // read topmost stack frame
        StackFrame before = mem.removeFrame();
        mem.addFrame( before );
        ControlFlow cf = null;
        
        // remember if this node has already been called
        if( mem.isDefined( "_call" + programPointer.getId(), Visibility.LOCAL ) )
        {
            String name = "_call" + programPointer.getId();
            mem.undeclare( name, Visibility.LOCAL );
            cf = programPointer.emptyForwardStep( mem );
            cf.setBackwardAction( (Memory m) -> {
                mem.declare( name, true, Visibility.LOCAL );
            });
        }
        else
            cf = programPointer.forwardStep( mem );
        // remember all ControlFlow actions that have been done to be able to undo it later
        controlStack.push( cf );
        
        // get the debug output
        currentDebugOutput = programPointer.getDebugOutput( mem );
        
        switch( cf.getStatus() )
        { // we are doing a "step into"
        case ControlFlow.STEP_INTO:
            // remember current line to be able to jump back to
            if( mem.isDefined( "_returnTo" + programPointer.getId(), Visibility.GLOBAL ) )
            {
                String name = "_returnTo" + programPointer.getId();
                mem.declare( name, mem.read( name, Visibility.GLOBAL ), Visibility.LOCAL );
                mem.undeclare( name, Visibility.GLOBAL );
                cf.setBackwardAction( (Memory m) -> {
                    mem.declare( name, mem.read( name, Visibility.LOCAL ), Visibility.GLOBAL );
                    mem.undeclare( name, Visibility.LOCAL );
                });
            }
            if( programPointer.getChildCount() == 0 )
                throw new IllegalStateException( "A Codeline without sublines tried to make a STEP_INTO." );
            else
                return selectNextNode( (PseudoCodeNode)programPointer.getFirstChild(), programPointer );
        case ControlFlow.STEP_OVER: // we are doing a "step over"
            if( programPointer.getParent() == null )
                return CodeStatus.FINISHED;
            // remember current line to be able to jump back to
            if( before.isDefined( "_returnTo" + programPointer.getId() ) )
            {
                String name = "_returnTo" + programPointer.getId();
                PseudoCodeNode nextPC = before.<PseudoCodeNode>get( name );
                before.undeclare( name );
                cf.setBackwardAction( (Memory m) -> {
                    before.declare( name, nextPC );
                });
                // continue with next node
                return selectNextNode( nextPC, programPointer );
            }
            // update program pointer
            PseudoCodeNode nextPC = (PseudoCodeNode) ((PseudoCodeNode)programPointer.getParent()).getChildAfter( programPointer );
            if( nextPC == null )
                return selectNextNode( (PseudoCodeNode)programPointer.getParent(), programPointer );
            else
                return selectNextNode( nextPC, programPointer );
        case ControlFlow.CALL: // we are doing a function call
            PseudoCodeNode f = cf.getFunction();
            // remember current line to be able to jump back to
            String name = "_call" + programPointer.getId();
            mem.declare( name, true, Visibility.LOCAL );
            mem.declare( "_returnTo" + f.getId(), cf.getJumpBack(), Visibility.GLOBAL );
            cf.setBackwardAction( (Memory m) -> {
                m.undeclare( "_returnTo" + f.getId(), Visibility.GLOBAL );
                m.undeclare( name, Visibility.LOCAL );
            });
            // go to next node
            return selectNextNode( f, programPointer );
        default:
            throw new IllegalStateException( "Unknown ControlFlow action" );
        }
    }
    
    private CodeStatus forwardStepOverUntilNotSkip() {
        // is the program pointer still available?
        if( programPointer == null )
            return CodeStatus.FINISHED;
        CodeStatus status = CodeStatus.UNFINISHED;
        // perform steps until someone sets skip to false
        do {
            insideNotSkipLoop = true;
            status = forwardStep();
            insideNotSkipLoop = false;
        } while( status == CodeStatus.UNFINISHED && skip );
        return status;
    }
    
    protected CodeStatus forwardStepOver()
    {
        // is the program pointer still available?
        if( programPointer == null )
            return CodeStatus.FINISHED;
        // remember stack size
        int stackSize = mem.getSize();
        CodeStatus status = CodeStatus.UNFINISHED;
        // perform steps until we arrive at the same stack size again
        do {
            status = forwardStep();
        } while( mem.getSize() > stackSize && status == CodeStatus.UNFINISHED );
        return status;
    }
    
    protected CodeStatus forwardStepOut()
    {
        // is the program pointer still available?
        if( programPointer == null )
            return CodeStatus.FINISHED;
        // remember stack size
        int stackSize = mem.getSize();
        CodeStatus status = CodeStatus.UNFINISHED;
        // perform steps until we arrive at a smaller stack size
        do {
            status = forwardStep();
        } while( mem.getSize() >= stackSize && status == CodeStatus.UNFINISHED );
        return status;
    }
    
    /**
     * returns the program pointer of the previous step
     * @return the node that was previously executed
     */
    protected PseudoCodeNode getLastProgramPointer() {
        if( controlStack.isEmpty() )
            return null;
        return controlStack.peek().getJumpBack();
    }
    
    protected CodeStatus backwardStep()
    {
        // check if we arrived at the beginning
        if( controlStack.isEmpty() )
            return CodeStatus.FINISHED;
        // what did we do before?
        ControlFlow cf = controlStack.pop();
        // where where we before?
        PseudoCodeNode nextPC = cf.getJumpBack();
        // undo action
        cf.backward( mem );
        nextPC.backwardStep( mem );
        
        currentDebugOutput = nextPC.getDebugOutput( mem );
        return selectBeforeNode( nextPC, programPointer );
    }
    
    private CodeStatus backwardStepOverUntilNotSkip()
    {
        // check if we arrived at the beginning
        if( programPointer == null )
            return CodeStatus.FINISHED;
        CodeStatus status = CodeStatus.UNFINISHED;
        // perform steps until someone sets skip to false
        do {
            insideNotSkipLoop = true;
            status = backwardStep();
            insideNotSkipLoop = false;
        } while( status == CodeStatus.UNFINISHED && skip );
        return status;
    }
    
    protected CodeStatus backwardStepOver()
    {
        // check if we arrived at the beginning
        if( programPointer == null )
            return CodeStatus.FINISHED;
        
        int stackSize = mem.getSize();
        CodeStatus status = CodeStatus.UNFINISHED;
        // perform steps until we arrrive at the same stack size again
        do {
            status = backwardStep();
        } while( mem.getSize() > stackSize && status == CodeStatus.UNFINISHED );
        return status;
    }
    
    protected CodeStatus backwardStepOut()
    {
        // check if we arrived at the beginning
        if( programPointer == null )
            return CodeStatus.FINISHED;
        
        int stackSize = mem.getSize();
        CodeStatus status = CodeStatus.UNFINISHED;
        // perform steps until we arrrive at a smaller stack size
        do {
            status = backwardStep();
        } while( mem.getSize() >= stackSize && status == CodeStatus.UNFINISHED );
        return status;
    }

    private synchronized void update()
    {
        if( renderImage )
        {
            renderImage = false;
            SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    for( ComponentListener l : view.getComponentListeners() )
                    {
                        l.componentResized( new ComponentEvent(view, 0) );
                        synchronized( this )
                        {
                            renderImage = true;
                        }
                    }
                }
            });
        }
    }
    
    @Override
    public void run()
    {
        while( true ) // if this loop would end we could not undo steps any more
        {
            CodeStatus status = null;
            try {
                Action action = controller.getNextAction(); // what action is next?
                graph.unselectGraph(); // remove selection from any nodes in the graph
                switch( action ) // perform the action
                {
                case FORWARD:
                    status = forwardStep();
                    break;
                case FORWARD_OUT:
                    status = forwardStepOut();
                    graph.unselectGraph();
                    break;
                case FORWARD_OVER:
                    status = forwardStepOver();
                    graph.unselectGraph();
                    break;
                case BACKWARD:
                    status = backwardStep();
                    break;
                case BACKWARD_OUT:
                    status = backwardStepOut();
                    graph.unselectGraph();
                    break;
                case BACKWARD_OVER:
                    status = backwardStepOver();
                    graph.unselectGraph();
                    break;
                default:
                    assert false;
                    break;
                }
            } catch (InterruptedException e) {
                e.printStackTrace();
                return;
            }
            
            // update the drawing
            update(); 
            
            if( status == CodeStatus.FINISHED ) // stop at the end
            {
                controller.setContinuous( false );
                controller.setNextAction( null );
            }
        }
    }
    
    public Memory getMemory()
    {
        return mem;
    }
    
    public String getDebugOutput()
    {
        return currentDebugOutput;
    }
}
